"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function typeToString(t) {
    switch (t.nodeType) {
        case "Named":
        case "Var":
            return t.name;
        case "Function":
            return `(${typeToString(t.from)} -> ${typeToString(t.to)})`;
    }
}
exports.typeToString = typeToString;
function unify(t1, t2) {
    if (t1.nodeType === "Named" &&
        t2.nodeType === "Named" &&
        t2.name === t1.name) {
        return {};
    }
    else if (t1.nodeType === "Var") {
        return varBind(t1.name, t2);
    }
    else if (t2.nodeType === "Var") {
        return varBind(t2.name, t1);
    }
    else if (t1.nodeType === "Function" && t2.nodeType === "Function") {
        const s1 = unify(t1.from, t2.from);
        const s2 = unify(applySubstToType(s1, t1.to), applySubstToType(s1, t2.to));
        return composeSubst(s1, s2);
    }
    else {
        throw `Type mismatch:\n    Expected ${typeToString(t1)}\n    Found ${typeToString(t2)}`;
    }
}
function varBind(name, t) {
    if (t.nodeType === "Var" && t.name === name) {
        return {};
    }
    else if (contains(t, name)) {
        throw `Type ${typeToString(t)} contains a reference to itself`;
    }
    else {
        const subst = {};
        subst[name] = t;
        return subst;
    }
}
function contains(t, name) {
    switch (t.nodeType) {
        case "Named":
            return false;
        case "Var":
            return t.name === name;
        case "Function":
            return contains(t.from, name) || contains(t.to, name);
    }
}
function composeSubst(s1, s2) {
    const result = {};
    for (const k in s2) {
        const type = s2[k];
        result[k] = applySubstToType(s1, type);
    }
    return Object.assign(Object.assign({}, s1), result);
}
function applySubstToType(subst, type) {
    switch (type.nodeType) {
        case "Named":
            return type;
        case "Var":
            if (subst[type.name]) {
                return subst[type.name];
            }
            else {
                return type;
            }
        case "Function":
            return {
                nodeType: "Function",
                from: applySubstToType(subst, type.from),
                to: applySubstToType(subst, type.to),
            };
    }
}
function addToContext(ctx, name, type) {
    const newEnv = Object.assign(Object.assign({}, ctx), { env: Object.assign({}, ctx.env) });
    newEnv.env[name] = type;
    return newEnv;
}
function newTVar(ctx) {
    const newVarNum = ctx.next;
    ctx.next++;
    return {
        nodeType: "Var",
        name: `T${newVarNum}`,
    };
}
function infer(ctx, e) {
    const env = ctx.env;
    switch (e.nodeType) {
        case "Int":
            return [{ nodeType: "Named", name: "Int" }, {}];
        case "Var":
            return inferVar(ctx, e);
        case "If":
            return inferIf(ctx, e);
        case "Function":
            return inferFunction(ctx, e);
        case "Call":
            return inferCall(ctx, e);
        case "Let":
            return inferLet(ctx, e);
    }
}
exports.infer = infer;
function inferVar(ctx, e) {
    const env = ctx.env;
    if (env[e.name]) {
        const envType = env[e.name];
        return [envType, {}];
    }
    else {
        throw `Unbound var ${e.name}`;
    }
}
function inferLet(ctx, expr) {
    const [rhsType, s1] = infer(ctx, expr.rhs);
    const ctx1 = applySubstToCtx(s1, ctx);
    const ctx2 = addToContext(ctx1, expr.name, rhsType);
    const [bodyType, s2] = infer(ctx2, expr.body);
    const s3 = Object.assign(Object.assign({}, s1), s2);
    return [bodyType, s3];
}
function inferFunction(ctx, e) {
    const newType = newTVar(ctx);
    const newCtx = addToContext(ctx, e.param, newType);
    const [bodyType, subst] = infer(newCtx, e.body);
    const inferredType = {
        nodeType: "Function",
        from: applySubstToType(subst, newType),
        to: bodyType,
    };
    return [inferredType, subst];
}
function inferIf(ctx, e) {
    const [condType, s0] = infer(ctx, e.cond);
    const s1 = unify(condType, {
        nodeType: "Named",
        name: "Bool",
    });
    const ctx1 = applySubstToCtx(composeSubst(s0, s1), ctx);
    const [_trueBranchType, s2] = infer(ctx1, e.trueBranch);
    const s3 = composeSubst(s1, s2);
    const ctx2 = applySubstToCtx(s2, ctx1);
    const [_falseBranchType, s4] = infer(ctx2, e.falseBranch);
    const s5 = composeSubst(s3, s4);
    const trueBranchType = applySubstToType(s5, _trueBranchType);
    const falseBranchType = applySubstToType(s5, _falseBranchType);
    const s6 = unify(trueBranchType, falseBranchType);
    return [applySubstToType(s6, trueBranchType), s6];
}
function inferCall(ctx, e) {
    const [funcType, s1] = infer(ctx, e.func);
    const [argType, s2] = infer(applySubstToCtx(s1, ctx), e.arg);
    const newVar = newTVar(ctx);
    const s3 = composeSubst(s1, s2);
    const s4 = unify({
        nodeType: "Function",
        from: argType,
        to: newVar,
    }, funcType);
    const s5 = composeSubst(s3, s4);
    const s6 = unify(applySubstToType(s5, funcType.from), argType);
    const resultSubst = composeSubst(s5, s6);
    return [applySubstToType(resultSubst, funcType.to), resultSubst];
}
function applySubstToCtx(subst, ctx) {
    const newContext = Object.assign(Object.assign({}, ctx), { env: Object.assign({}, ctx.env) });
    for (const name in newContext.env) {
        const t = newContext.env[name];
        newContext.env[name] = applySubstToType(subst, t);
    }
    return newContext;
}
//# sourceMappingURL=typeInference.js.map